local M = {}

local stringFormatter = require("scenario/gravitationalRacing/utils/stringFormatting")
local ClassVector     = require("scenario/gravitationalRacing/classes/classVector")
local ClassCelestial  = require("scenario/gravitationalRacing/classes/classCelestial")
local factors         = require("scenario/gravitationalRacing/dataValues/factors")
local field           = require("scenario/gravitationalRacing/celestial/gravitationalScalarField")
local tableComp       = require("scenario/gravitationalRacing/utils/tableComprehension")

local celestials = {fixedDynamic = {}, dynamic = {}, static = {}}
local initial = {positions = {length = 0}, scale = {}}
local passiveMode = false
local spectating, camera = nil, nil

local function findCelestial(name, objectType)
	--[[
	Returns the celestial object in the table of celestials in the scenario
	This function MUST be given the celestial name, but can also receieve an optional
	parameter to narrow the search (object type)
	]]--
	if objectType then
		return celestials[objectType][name]
	else
		for type, table in pairs(celestials) do
			if table[name] then
				return table[name]
			end
		end
	end

	log(I, "gravitationalRacing: findCelestial()", "Could not find celestial with name="..name..", ignoring...")
end

local function findNextName(type)
	--[[
	Finds the name of the next object (ie. if there are two black holes in-world, returns blackhole_3)
	]]--
	if not type then
		return
	end

	local i = 1
	local name = type.."_"..i
	while scenetree.findObject(name) do
		i = i + 1
		name = type.."_"..i
	end

	return name
end

local function findCelestialMatch(func, value, returnInstances)
	--[[
	Finds ALL celestials whose particular method return value equals value
	returnInstances is a n optional parameter that tells this function to return instance,
	not names instead
	TODO allow arguments
	]]--
	local matching = {}

	for _, objType in pairs(celestials) do
		for _, instance in pairs(objType) do
			local result = instance:callMethod(func)
			if type(result) == "table" then
				if result["instanceOf"] and result:instanceOf() == "ClassVector" then
					--TODO This should probably be more generic
					if result:isEqualTolerance(value, 10) then
						table.insert(matching, returnInstances and instance or instance:getName())
					end
				else
					log("E", "findCelestialMatch()", "I don't know how to compare this table!")
				end
			elseif result == value then
				table.insert(matching, returnInstances and instance or instance:getName())
			end
		end
	end

	return matching
end

local function callMethodOnCelestials(list, method, args)
	--[[
	Calls a method on a set of celestials, or all if not specified
	]]--
	local check = list ~= nil

	for _, objType in pairs(celestials) do
		for _, instance in pairs(objType) do
			if check then
				local instName = instance:getName()
				for _, name in ipairs(list) do
					if name == instName then
						instance:callMethod(method, args)
						break
					end
				end
			else
				instance:callMethod(method, args)
			end
		end
	end
end

local function getOrbitalVelocity(celestial, parentCelestial)
	--[[
	Calculates the orbital velocity of an object, given an instance of the parent
	celestial body
	Note: Formula derived from F = GMm/r² = mv²/r, --> v = sqrt(GM/r) only works for NON-STARS
	Another formula can be derived for stars: a = F/m, F = GMm/r², a = v²/r -->  v = sqrt(GM/4r)
	Note2: The vector returned will always have the object going anti-clockwise
	around its parent celestials, and starts in the +y direction
	]]--
	local distance = parentCelestial:getPosition():subtract(celestial:getPosition())
	local r = distance:getMagnitude() / factors.getDistanceScaleFactor()

	local celestialType = celestial:getType()

	--Since this binary version of the orbital velocity formula requires m1 ≈ m2, stars can orbit black holes as m1 >> m2 - for all intents and purposes
	if ((celestialType == "star" and parentType == "star") or (celestialType == "blackhole" and parentType == "blackhole")) then
		return ClassVector.new(0, math.sqrt(ClassCelestial.getGravitationalConstant() * parentCelestial:getMass()/(4*r)) * factors.getDistanceScaleFactor() * factors.getTimeScaleFactor(), 0)
	else
		return ClassVector.new(0, math.sqrt(ClassCelestial.getGravitationalConstant() * parentCelestial:getMass()/r) * factors.getDistanceScaleFactor() * factors.getTimeScaleFactor(), 0)
	end
end

local function setInitialVelocities()
	--[[
	Sets the initial velocites for all dynamic celestials
	]]--
	local setVel = function(celestial, orbitingBodyName, direction)
		--[[
		Gets and sets the orbital velocity of a celestial given its parent celestial
		]]--
		local stellarCelestial = findCelestial(orbitingBodyName)
		v = getOrbitalVelocity(celestial, stellarCelestial)

		if direction == "neg" then
			v = v:multiply(-1)
		end

		celestial:setVelocity(v)

		--Don't override the path already set
		if not celestial:hasPlacedPath() then
			celestial:setupPath(stellarCelestial)
		end
	end

	local stars = {}
	-- Set the velocity of non-stars, and store stars in another table to handle later
	for k, celestial in pairs(tableComp.merge(celestials.dynamic, celestials.fixedDynamic)) do
		if ((celestial:getType() ~= "star" and celestial:getType() ~= "blackhole") or celestial:getObjectType() == "fixedDynamic") and celestial:getParentCelestial() then
			setVel(celestial, celestial:getParentCelestial(), "pos")
		elseif celestial:getObjectType() == "fixedDynamic" then
			if not celestial:hasPlacedPath() then
				celestial:setupPath()
			end
		elseif celestial:getType() == "star" then
			--Zero velocity for stars
			celestial:setVelocity(ClassVector.new(0, 0, 0))
		end
	end

	-- --Handle stars' velocities--
	-- if #stars > 2 then
	--
	-- end
	--
	--
	-- local names = {stars[1]:getName(), stars[2]:getName()}
	-- local directions = {"pos", "neg"}
	-- --To make sure they go ani-clockwise, the star who has a larger position Y
	-- --must be the one to recieve a positive acceleration in the X axis
	-- local star1InFront = stars[1]:getPosition():getY() > stars[2]:getPosition():getY()
	--
	-- for i, star in ipairs(stars) do
	-- 	name = names[#names+1-i]
	-- 	if star1InFront then
	-- 		--Sets the first ot be positive, the second to be negative
	-- 		setVel(star, name, directions[i])
	-- 	else
	-- 		--Sets the first ot be negative, the second to be positive
	-- 		setVel(star, name, directions[#directions+1-i])
	-- 	end
	-- end
end

local function setInitialConditions()
	--[[
	Sets the initial starting position and scale for each celestial
	If the information does not exist, create it and stored it
	]]--
	--Only one length variable needs to be stored for both positions and scale tables,
	--as they will have the same value
	if initial.positions.length == 0 then
		--Save initial conditions (indicates the scenario has been reloaded or first time loaded)
		for type, table in pairs(celestials) do
			initial.positions[type] = {}
			initial.scale[type] = {}
			for k, celestial in pairs(table) do
				local obj = scenetree.findObject(k)

				local objPos = obj:getPosition()
				initial.positions[type][k] = ClassVector.new(objPos.x, objPos.y, objPos.z)
				initial.positions.length = initial.positions.length + 1

				initial.scale[type][k] = obj:getScale().x
			end
		end
	end

	--Set object's position in its instance
	for type, table in pairs(celestials) do
		for k, celestial in pairs(table) do
			celestial:setSetup(false)

			local pos = initial.positions[type][k]
			local initPos = ClassVector.new(pos:getX(), pos:getY(), pos:getZ())

			celestial:setPosition(initPos)
			celestial:setInitialPosition(initPos)

			local scale = initial.scale[type][k]
			celestial:setScale(scale)

			local newRadius = scale*factors.getRadiusScaleFactor()
			celestial:setRadius(newRadius)

			local newMass = celestial:calculateMass(newRadius, celestial:getActual("radius"), celestial:getActual("mass"))
			celestial:setMass(newMass)

			--Update its position and size in world
			celestial:display()
		end
	end
end

local function findCelestials()
	--[[
	Finds all celestials in the scenario
	]]--
	local celestialsInScenario = {}
	local categories = ClassCelestial.getTypes()

	local i, obj
	for _, category in ipairs(categories) do
		i = 1
		obj = scenetree.findObject(category.."_"..i)
		--While they exist, add them into the table
		while obj do
			table.insert(celestialsInScenario, {category = category, obj = obj})
			i = i + 1
			obj = scenetree.findObject(category.."_"..i)
		end
	end

	return celestialsInScenario
end

local function createCelestials(fullReset, vehicles, passive)
	--[[
	Creates the celestials in the scenario into objects
	]]--
	passiveMode = passive

	--Stores all supernova objects, as they need to survey the area around them
	local supernovae = {}

	--When loading another scenario, the data here is still stored, so can be used for setting up
	if fullReset then
		--Reset/update variables
		celestials = {fixedDynamic = {}, dynamic = {}, static = {}}
		initial = {positions = {length = 0}, scale = {}}

		local foundCelestials = findCelestials()
		--Stores all children to a parent
		local children = {}

		for _, celestial in ipairs(foundCelestials) do
			local objData = celestial.obj

			local name = objData.name
			local objectType = objData.objectType

			if not objectType or (objectType == "fixedDynamic" and not objData.path) then
				error("Object "..name.." has not been properly setup as a celestial! [objectType="..(objectType or "nil")..", path="..(objData.path or "nil").."]")
			end

			local celestialType = name:gsub("_%d+", "")
			local class = objData.celestialClass
			local parent = objData.orbitingBody

			local isPassive
			--The celestial gets preferential treatment over passivity
			if objData.passive then
				if tonumber(objData.passive) == 1 then
					passive = true
				else
					passive = false or passiveMode
				end
			else
				isPassive = passiveMode
			end

			local instance
			if not class or class == "stable" then
				instance = ClassCelestial.new         (name, celestialType, objData:getScale().x, objectType, parent, isPassive, objData.delayed or false, objData.axisTilt)
			elseif class == "unstable" then
				instance = ClassCelestial.newUnstable (name, celestialType, objData:getScale().x, objectType, parent, isPassive, objData.delayed or false, objData.axisTilt, objData.ferocity)
			elseif class == "exotic" then
				instance = ClassCelestial.newExotic   (name, celestialType, objData:getScale().x, objectType, parent, isPassive, objData.delayed or false, objData.axisTilt)
			elseif class == "supernova" then
				local supernovaData = {
					vehicles = vehicles,
				}

				instance = ClassCelestial.newSupernova(name, celestialType, objData:getScale().x, objectType, parent, passiveMode, objData.delayed or false, objData.axisTilt, supernovaData)
				table.insert(supernovae, instance)
			end

			celestials[objectType][name] = instance

			--Add in children to parents
			if parent then
				if not children[parent] then
					children[parent] = {instance}
				else
					table.insert(children[parent], instance)
				end
			end
		end

		--Add a reference to each parent from the children
		for parent, childInstances in pairs(children) do
			local parentInstace = findCelestial(parent)
			for _, childInstance in ipairs(childInstances) do
				parentInstace:addChild(childInstance)
			end
		end
	else
		--Setup danger zone areas and delete old trails
		for type, groupOfCelestials in pairs(celestials) do
			for k, value in pairs(groupOfCelestials) do
				local instance = celestials[type][k]

				if not passiveMode then
					instance:displayRadialZones()
				end

				instance:reset()
			end
		end
	end

	setInitialConditions()
	--Note: velocities need to be set after as to calculate this the positions
	--need to be setup beforehand
	setInitialVelocities()

	--Needs to happen after above functions as the celestials need to be positioned first
	if fullReset then
		--Add a reference to any celestial that will be affected by a supernova explosion
		for sName, supernova in pairs(supernovae) do
			local sPosition = supernova:getPosition()
			local effectRange = supernova:getEffectRange()

			local celestialsInRange = {}
			for cName, instance in pairs(tableComp.merge(celestials.dynamic, tableComp.merge(celestials.fixedDynamic, celestials.static))) do
				--Don't reference itself and a supernova only affects planets
				if sName ~= cName and instance:getType() == "planet" then
					if sPosition:getDistanceBetween(instance:getPosition()) <= effectRange then
						table.insert(celestialsInRange, instance)
					end
				end
			end

			supernova:setCelestialsWithinRange(celestialsInRange)
		end
	end
end

local function update(start, vehicles, dt)
	--[[
	Updates all celestials in the scenario
	]]--
	if bullettime.get() > 0 then
		local attractVehicles = function(celestial, invert)
			--[[
			Attracts the player car to the celestial
			]]--
			--Even if the scenario/event is non-passive, a celestial can still be individually
			if not celestial:getPassivity() then
				for _, vehInstance in pairs(vehicles) do
					if not vehInstance:isImmune() then
						local forceOnPlayer = celestial:getForce(vehInstance)
						if invert then
							forceOnPlayer = forceOnPlayer:multiply(-1)
						end

						vehInstance:addWind(forceOnPlayer)
						celestial:handleCollision(vehInstance)
					end
				end
			end
		end

		local actualDt = dt * bullettime.get()

		--Dynamic celestials wait until the scenario start
		if start then
			for k, celestial1 in pairs(celestials.dynamic) do
				if not celestial1:isRemoved() then
					--Account for the possibility that the object has been destroyed previously
					--by another object
					if not celestial1:isMarkedForDestruction() then
						for k2, celestial2 in pairs(celestials.dynamic) do
							--Do not attract object to itself
							if k ~= k2 then
								if not celestial2:isMarkedForDestruction() and not celestial2:isRemoved() then
									local force = celestial1:getForce(celestial2)
									celestial1:applyForce(force, actualDt)
									celestial1:handleCollision(celestial2)
								end
							end
						end
					end

					if not passiveMode then
						--Affect the player
						attractVehicles(celestial1)
					end

					celestial1:update(actualDt)
				end
			end
		end

		for k, celestial in pairs(celestials.fixedDynamic) do
			if not celestial:isRemoved() then
				if start and not passiveMode then
					attractVehicles(celestial, true)
				end

				celestial:update(actualDt)
			-- else
				--TODO - check if commenting this out causes problems with Celestial Creator UI
				-- --Remove reference
				-- celestials.fixedDynamic[k] = nil
			end
		end

		for k, celestial in pairs(celestials.static) do
			if not celestial:isRemoved() then
				if start and not passiveMode then
					attractVehicles(celestial, true)
				end

				celestial:update(actualDt)
			-- else
				-- --Remove reference
				-- celestials.static[k] = nil
			end
		end

		if start then
			for _, vehInstance in pairs(vehicles) do
				vehInstance:update(actualDt)
			end
		end

		-- if spectating then
		-- 	local pos = spectating:getPosition()
		-- 	local radius = spectating:getScaledRadius()
		-- 	camera:setPosition(vec3(pos:getX(), pos:getY(), pos:getZ() + 250):toPoint3F())
		-- end
	end
end

local function createCelestial(shapeName, initPos, name, type, scale, objectType, orbitingBody, passive, delayed, rotation, pathData)
	--[[
	Creates a new celestial
	]]--
	--Create the object
	if not scenetree.findObject(name) then
		if objectType == "static" then
			TorqueScript.eval([[
			new TSStatic(]]..(name)..[[) {
				shapeName = "levels/smallgrid/art/gravitationalRacing/celestialbodies/]]..shapeName..[[.dae";
				dynamic = "1";
				scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
				objectType = "static";
				position = "]]..initPos:getX()..[[ ]]..initPos:getY()..[[ ]]..initPos:getZ()..[[";
				collisionType = "Visible Mesh Final";
				decalType = "Visible Mesh Final";
			};
			]])
		else
			TorqueScript.eval([[
			new TSStatic(]]..(name)..[[) {
				shapeName = "levels/smallgrid/art/gravitationalRacing/celestialbodies/]]..shapeName..[[.dae";
				dynamic = "1";
				scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
				objectType = ]]..objectType..[[;
				position = "]]..initPos:getX()..[[ ]]..initPos:getY()..[[ ]]..initPos:getZ()..[[";
			};
			]])
		end
	end

	if type == "blackhole" then
		if not scenetree.findObject(name.."_disc") then
			TorqueScript.eval([[
			new TSStatic(]]..(name)..[[_disc) {
				shapeName = "levels/smallgrid/art/gravitationalRacing/celestialbodies/blackhole_discJets.dae";
				dynamic = "1";
				scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
				position = "]]..initPos:getX()..[[ ]]..initPos:getY()..[[ ]]..initPos:getZ()..[[";
			};
			]])
		end
	end

	if orbitingBody then
		TorqueScript.eval(name..'.orbitingBody = "'..orbitingBody..'";')
	end

	if delayed then
		TorqueScript.eval(name..'.delayed = "'..tostring(delayed)..'";')
	end

	if pathData then
		for k, v in pairs(pathData) do
			TorqueScript.eval(name..'.'..k..' = "'..v..'";')
		end
	end

	--Create the associated instance
	local obj = ClassCelestial.new(name, type, scale, objectType, orbitingBody, passive, delayed, rotation)
	obj:setPosition(initPos)
	--Add it into the table for updating
	celestials[objectType][name] = obj

	--Return a reference for use in another file
	return obj
end

local function deleteCelestial(name, objectType)
	--[[
	Deletes a celestial
	(objectType is optional for quicker deletion)
	]]--
	if objectType then
		local instance = celestials[objectType][name]
		instance:delete()

		celestials[objectType][name] = nil
	else
		for _, group in ipairs(celestials) do
			local instance = celestials[group][name]
			if instance then
				instance:delete()
				celestials[group][name] = nil
			end
		end
	end
end

local function setSpectate(name, objectType)
	--[[
	Spectates an object
	NOTE: Experimental
	]]--
	spectating = findCelestial(name, objectType)
	commands.setFreeCamera()
	camera = commands.getCamera(commands.getGame())
end

local function createForceField()
	--[[
	Creates a force scale field, showing the strength of the gravitational force
	over an area
	]]--
	field.create(celestials)
end

local function removeAllCelestials()
	celestials = {fixedDynamic = {}, dynamic = {}, static = {}}
end

local function printResults()
	--[[
	A testing function for printing the various celestial related data it has created
	]]--
	print("Celestials: [Passive Mode = "..tostring(passiveMode).."]")
	print()
	local celestialInfo = {}
	local keys = ClassCelestial.toListKeys()

	for k, t in pairs(celestials) do
		for k2, instance in pairs(t) do
			table.insert(celestialInfo, instance:toList())
		end
	end

	stringFormatter.printClassInfo(celestialInfo, keys)

	print()
	print("Scenario running with factors:")
	print("  Distance Factor: "..factors.getDistanceScaleFactor())
	print("  Time Factor: "..factors.getTimeScaleFactor())
	print("  Radius Factor: "..factors.getActualRadiusScaleFactor())
	print()
end

M.findCelestial = findCelestial
M.findNextName = findNextName
M.findCelestialMatch = findCelestialMatch
M.callMethodOnCelestials = callMethodOnCelestials
M.getOrbitalVelocity = getOrbitalVelocity
-- M.enablePassiveMode = enablePassiveMode
-- M.disablePassiveMode = disablePassiveMode
M.createCelestials = createCelestials
M.update = update
M.createForceField = createForceField
M.createCelestial = createCelestial
M.deleteCelestial = deleteCelestial
M.setSpectate = setSpectate
M.removeAllCelestials = removeAllCelestials
M.printResults = printResults
return M
